import * as Electron from 'electron';

import { ObjectUtil } from '@ucap/core';

import { EventHandlerRegistry } from './registries/event-handler.registry';
import { EventAttachType } from './types/event-attach.type';
import { EventHandlerMetadata } from './models/event-handler.metadata';

export abstract class Attacher {
  private static readonly eventHandlerMap: Map<
    any,
    { channel: string; listener: (...args: any[]) => void }[]
  > = new Map();

  static attach(instance: any) {
    const metadatas = EventHandlerRegistry.getEventHandlers(
      ObjectUtil.classOf(instance)
    );

    if (!!metadatas) {
      metadatas.reverse().forEach((metadata) => {
        metadata.eventHandlers.forEach((channels, attachType) => {
          switch (attachType) {
            case EventAttachType.on:
              {
                channels.forEach((channel) => {
                  const listener = (...args: any[]) => {
                    Attacher.invoke(metadata, instance, args);
                  };

                  Electron.app.on(channel as any, listener);
                  Attacher.addEventHandler(instance, channel, listener);
                });
              }
              break;
          }
        });
      });
    }
  }

  static dettach(instance: any) {
    Attacher.removeEventHandler(instance);
  }

  static dettachAll() {
    Electron.app.removeAllListeners();
  }

  static invoke(
    metadata: EventHandlerMetadata,
    instance: any,
    args: any[]
  ): any {
    if (!!metadata.parent) {
      return Attacher.invoke(metadata.parent, instance, args);
    }

    // tslint:disable-next-line: ban-types
    return (metadata.target.prototype[metadata.propertyKey] as Function).call(
      instance,
      ...args
    );
  }

  private static addEventHandler(instance: any, channel: any, listener: any) {
    let listeners = Attacher.eventHandlerMap.get(instance);
    if (!listeners) {
      listeners = [];
      Attacher.eventHandlerMap.set(instance, listeners);
    }

    listeners.push({ channel, listener });
  }

  private static removeEventHandler(instance: any) {
    const listeners = Attacher.eventHandlerMap.get(instance);
    if (!listeners) {
      return;
    }

    for (const listener of listeners) {
      Electron.app.removeListener(listener.channel as any, listener.listener);
    }
  }
}
