import { Type, ReflectUtil, ObjectUtil } from '@ucap/core';

import { AutoUpdaterChannel } from '@ucap/electron-core';

import { EventHandlerMetadata } from '../models/event-handler.metadata';
import { EventAttachType } from '../types/event-attach.type';

export const REFLECT_KEY_EVENTHANDLERS =
  'ucap_electron_auto_updater_event_handlers';

export class EventHandlerRegistry {
  static getOwnEventHandlers(target: Type<any>) {
    if (!this.hasEventHandlers(target)) {
      ReflectUtil.set(REFLECT_KEY_EVENTHANDLERS, [], target);
    }

    return ReflectUtil.getOwn(REFLECT_KEY_EVENTHANDLERS, target);
  }

  static getEventHandlers(target: Type<any>): EventHandlerMetadata[] {
    return this.getOwnEventHandlers(target).concat(this.inherit(target));
  }

  static hasEventHandlers(target: Type<any>): boolean {
    return ReflectUtil.hasOwn(REFLECT_KEY_EVENTHANDLERS, target);
  }

  static get(
    target: Type<any>,
    propertyKey: string | symbol
  ): EventHandlerMetadata {
    if (!this.has(target, propertyKey)) {
      const evnetHandler = new EventHandlerMetadata({
        target,
        propertyKey
      });
      this.getOwnEventHandlers(target).push(evnetHandler);
      ReflectUtil.set(
        REFLECT_KEY_EVENTHANDLERS,
        evnetHandler,
        target,
        propertyKey
      );
    }

    return ReflectUtil.getOwn(REFLECT_KEY_EVENTHANDLERS, target, propertyKey);
  }

  static has(target: Type<any>, method: string | symbol): boolean {
    return ReflectUtil.hasOwn(REFLECT_KEY_EVENTHANDLERS, target, method);
  }

  static use(
    target: Type<any>,
    targetKey: string | symbol,
    attachType: EventAttachType,
    channel: AutoUpdaterChannel
  ) {
    this.get(target, targetKey).merge(attachType, channel);

    return this;
  }

  private static inherit(clazz: Type<any>) {
    const eventHandlers: EventHandlerMetadata[] = [];
    let inheritedClass = ObjectUtil.getInheritedClass(clazz);

    while (
      inheritedClass &&
      EventHandlerRegistry.hasEventHandlers(inheritedClass)
    ) {
      this.getOwnEventHandlers(inheritedClass).forEach(
        (eventHandler: EventHandlerMetadata) => {
          eventHandlers.push(inheritEndpoint(clazz, eventHandler));
        }
      );

      inheritedClass = ObjectUtil.getInheritedClass(inheritedClass);
    }

    return eventHandlers;
  }
}

function inheritEndpoint(
  target: Type<any>,
  eventHandler: EventHandlerMetadata
): EventHandlerMetadata {
  return new EventHandlerMetadata({
    ...eventHandler,
    target,
    type: eventHandler.type,
    parent: eventHandler
  });
}
