import { Type, Metadata, ReflectUtil, ObjectUtil, Store } from '@ucap/core';

import { BrowserWindowChannel } from '@ucap/electron-core';

import { EventAttachType } from '../types/event-attach.type';

export interface EventHandlerConstructorOptions {
  target: Type<any>;
  propertyKey: string | symbol;
  eventHandlers?: Map<EventAttachType, BrowserWindowChannel[]>;
  type?: any;
  parent?: EventHandlerMetadata;
}

export class EventHandlerMetadata extends Metadata
  implements EventHandlerConstructorOptions {
  eventHandlers: Map<EventAttachType, BrowserWindowChannel[]> = new Map();

  readonly parent: EventHandlerMetadata | undefined;

  constructor(options: EventHandlerConstructorOptions) {
    super(
      options.target,
      options.propertyKey,
      Object.getOwnPropertyDescriptor(options.target, options.propertyKey)
    );
    const {
      target,
      parent,
      propertyKey,
      eventHandlers = new Map(),
      type
    } = options;

    this._type = ReflectUtil.getReturnType(target, propertyKey);

    this.eventHandlers = eventHandlers;
    this.type = type;
    this.parent = parent;
  }

  get type(): Type<any> {
    return ObjectUtil.isPromise(this._type) ||
      ObjectUtil.isArrayOrArrayClass(this._type) ||
      this._type === Object
      ? undefined!
      : this._type;
  }

  set type(type: Type<any>) {
    this._type = type;
  }

  get(key: any) {
    const ctrlValue = Store.from(this.target).get(key);
    let meta = ObjectUtil.deepExtends(undefined, ctrlValue);
    const endpointValue = this.store.get(key);
    if (endpointValue !== undefined) {
      meta = ObjectUtil.deepExtends(meta, endpointValue);
    }

    return meta;
  }

  merge(attachType: EventAttachType, channel: BrowserWindowChannel): this {
    if (!this.eventHandlers.has(attachType)) {
      this.eventHandlers.set(attachType, []);
    }

    this.eventHandlers.get(attachType).push(channel);

    return this;
  }
}
